@@ -56,6 +56,32 @@
 #include "io_lib/bam.h"
 #include "io_lib/os.h"
 #include "io_lib/thread_pool.h"
+#include "io_lib/crc32.h"
+#include "io_lib/bgzip.h"
+
+// On later gcc releases the ALLOW_UAC code causes the vectorizor to
+// use aligned SIMD instructions on unaligned memory access.  This is due
+// to our own abuse of char to int aliasing, but doing things the legal
+// way is still slower overall (5-14% depending on system and compiler).
+// However by explicitly altering the alignment of the integer types
+// we can persuade the compiler to generate the unaligned SIMD
+// instructions instead.
+#if defined(__GNUC__) && !(defined(__clang__) || defined(__ICC))
+typedef  int16_t  int16_u __attribute__ ((aligned (1)));
+typedef uint16_t uint16_u __attribute__ ((aligned (1)));
+typedef  int32_t  int32_u __attribute__ ((aligned (1)));
+typedef uint32_t uint32_u __attribute__ ((aligned (1)));
+#else
+typedef  int16_t  int16_u;
+typedef uint16_t uint16_u;
+typedef  int32_t  int32_u;
+typedef uint32_t uint32_u;
+#endif
+
+// If using Cloudflare's zlib, consider just switching to the zlib crc.
+// However this doesn't work on older machines.
+
+//#define iolib_crc32 crc32
 
 #define USE_MT
 
@@ -101,6 +127,7 @@
 static int bam_more_input(bam_file_t *b);
 static int bam_uncompress_input(bam_file_t *b);
 static int reg2bin(int start, int end);
+static int bgzf_block_write(bam_file_t *bf, int level, const void *buf, size_t count);
 static int bgzf_write(bam_file_t *bf, int level, const void *buf, size_t count);
 static int bgzf_write_mt(bam_file_t *bf, int level, const void *buf, size_t count);
 #ifdef USE_MT
@@ -409,6 +436,11 @@ static void bam_file_init(bam_file_t *b) {
     b->eof      = 0;
     b->nd_jobs    = 0;
     b->ne_jobs    = 0;
+    b->idx = NULL;
+    b->current_block = 0;
+    b->bgbuf_p = b->bgbuf;
+    b->bgbuf_sz = 0;
+    b->idx_fn = NULL;
 }
 
 /*! Opens a SAM or BAM file.
@@ -470,6 +502,9 @@ bam_file_t *bam_open(const char *fn, const char *mode) {
 	    goto error;
     }
 
+    if (NULL == (b->idx = gzi_index_init()))
+      goto error;
+
     /* Load first block so we can check */
     bam_more_input(b);
     if (b->comp_sz >= 2 && b->comp_p[0] == 31 && b->comp_p[1] == 139)
@@ -550,7 +585,8 @@ int bam_close(bam_file_t *b) {
 
     if (b->mode & O_WRONLY) {
 	if (b->binary) {
-	    if (BGZF_WRITE(b, b->level, b->uncomp, b->uncomp_p - b->uncomp)) {
+	    if (bgzf_block_write(b, b->level, b->uncomp,
+				 b->uncomp_p - b->uncomp)) {
 		fprintf(stderr, "Write failed in bam_close()\n");
 	    }
 
@@ -572,6 +608,7 @@ int bam_close(bam_file_t *b) {
 
     if (b->bs)
 	free(b->bs);
+
     if (b->header)
 	sam_hdr_free(b->header);
 
@@ -584,6 +621,13 @@ int bam_close(bam_file_t *b) {
     if (b->fp)
 	r = fclose(b->fp);
 
+    if (b->idx) {
+	if ((b->mode == O_RDONLY) && b->idx_fn) {
+	    gzi_index_dump(b->idx, b->idx_fn, NULL);
+	}
+	gzi_index_free(b->idx);
+    }
+
     if (b->pool) {
 	/* Should be no BAM jobs left in the pool, but if we abort on
 	 * and error and close early then we need to drain the pool of
@@ -639,6 +683,7 @@ typedef struct {
     unsigned char comp[Z_BUFF_SIZE];
     unsigned char uncomp[Z_BUFF_SIZE];
     size_t comp_sz, uncomp_sz;
+    int ignore_chksum;
 } bgzf_decode_job;
 static bgzf_decode_job *last_job = NULL;
 
@@ -668,6 +713,18 @@ void *bgzf_decode_thread(void *arg) {
 	return NULL;
     }
 
+    if (!j->ignore_chksum) {
+	uint32_t crc1=iolib_crc32(0L, (unsigned char *)j->uncomp, s.total_out);
+	uint32_t crc2;
+	memcpy(&crc2, j->comp + j->comp_sz, 4);
+	crc2 = le_int2(crc2);
+	if (crc1 != crc2) {
+	    fprintf(stderr, "Invalid CRC in Deflate stream: %08x vs %08x\n",
+		    crc1, crc2);
+	    return NULL;
+	}
+    }
+
     j->uncomp_sz  = s.total_out;
 
     return j;
@@ -788,8 +845,9 @@ static int bam_uncompress_input(bam_file_t *b) {
 		    } while (b->comp_sz < bsize + 8);
 		}
 
-		memcpy(j->comp, b->comp_p, bsize);
-		j->comp_sz = bsize+1;
+		memcpy(j->comp, b->comp_p, bsize+8);
+		j->comp_sz = bsize;
+		j->ignore_chksum = b->ignore_chksum;
 
 		b->comp_p  += bsize + 8; // crc & isize
 		b->comp_sz -= bsize + 8; // crc & isize
@@ -843,7 +901,10 @@ static int bam_uncompress_input(bam_file_t *b) {
 #endif
 	b->uncomp_sz = j->uncomp_sz;
 	t_pool_delete_result(res, 0);
-
+	if (b->idx){
+	    if (gzi_index_add_block(b->idx, j->comp_sz + 26, b->uncomp_sz))
+		return -1;
+	}
     } else {
 	/* Single threaded version, or non-bgzf format data */
 
@@ -907,7 +968,7 @@ static int bam_uncompress_input(bam_file_t *b) {
 		    }
 		} while (b->comp_sz < bsize + 8);
 	    }
-	    
+
 	    b->s.avail_in  = bsize;
 	    b->s.next_in   = b->comp_p;
 	    b->s.avail_out = Z_BUFF_SIZE;
@@ -927,8 +988,27 @@ static int bam_uncompress_input(bam_file_t *b) {
 	    b->comp_sz  -= bsize + 8;
 	    b->uncomp_sz  = b->s.total_out;
 	    b->uncomp_p   = b->uncomp;
+
+	    if (b->idx){
+		if (gzi_index_add_block(b->idx, bsize + 26, b->uncomp_sz))
+		    return -1;
+	    }
+
+	    if (!b->ignore_chksum) {
+		uint32_t crc1 = iolib_crc32(0L, (unsigned char *)b->uncomp,
+					    b->uncomp_sz);
+		uint32_t crc2;
+		memcpy(&crc2, b->comp_p-8, 4);
+		crc2 = le_int2(crc2);
+		if (crc1 != crc2) {
+		    fprintf(stderr, "Invalid CRC in Deflate stream: "
+			    "%08x vs %08x\n", crc1, crc2);
+		    return -1;
+		}
+	    }
 	} else {
 	    /* Some other gzip variant, but possibly still having xlen */
+	    /* NB: we don't check CRCs here, but I don't think these BAMs exist either */
 	    while (xlen) {
 		int d = MIN(b->comp_sz, xlen);
 		xlen     -= d;
@@ -959,6 +1039,7 @@ static int bam_uncompress_input(bam_file_t *b) {
 
 		if (err == Z_STREAM_END) {
 		    b->z_finish = 1;
+
 		    /* Consume (ignore) CRC & ISIZE */
 		    if (b->comp_sz < 8)
 			bam_more_input(b);
@@ -1256,7 +1337,8 @@ static int sam_next_seq(bam_file_t *b, bam_seq_t **bsp) {
     }
     bam_set_cigar_len(bs, cigar_len);
     //printf("pos %d, %d..%d => bin %d\n", bs->pos, start, end, reg2bin(start, end));
-    bam_set_bin(bs, reg2bin(start,end));
+    // SAM spec changed 8th April 2014 (f7651377) for unmapped data.
+    bam_set_bin(bs, reg2bin(start, bs->flag & BAM_FUNMAP ? start+1 : end));
     if (!*cpf++) return -1;
     
     /* mate ref name */
@@ -1985,8 +2067,10 @@ int bam_aux_iter(bam_seq_t *b, char **iter_handle,
     char k3[3];
     int r = bam_aux_iter_full(b, iter_handle, k3, type, val);
 
-    key[0] = k3[0];
-    key[1] = k3[1];
+    if (r == 0) {
+	key[0] = k3[0];
+	key[1] = k3[1];
+    }
 
     return r;
 }
@@ -2827,8 +2911,8 @@ static int bgzf_encode(int level,
     blk[16] = ((cdata_size + 25) >> 0) & 0xff;
     blk[17] = ((cdata_size + 25) >> 8) & 0xff;
 
-    crc = crc32(0L, NULL, 0L);
-    crc = crc32(crc, (unsigned char *)buf, in_sz);
+    crc = iolib_crc32(0L, NULL, 0L);
+    crc = iolib_crc32(crc, (unsigned char *)buf, in_sz);
     blk[18+cdata_size+0] = (crc >> 0) & 0xff;
     blk[18+cdata_size+1] = (crc >> 8) & 0xff;
     blk[18+cdata_size+2] = (crc >>16) & 0xff;
@@ -2843,6 +2927,44 @@ static int bgzf_encode(int level,
     return 0;
 }
 
+static int bgzf_block_write(bam_file_t *bf, int level,
+			    const void *buf, size_t count) {
+    if (!bf->idx)
+	return BGZF_WRITE(bf, level, buf, count);
+
+    const uint8_t *input = (const uint8_t*)buf;
+    // amount of uncompressed data to be fed into next block
+    uint64_t ublock_size;
+    ssize_t remaining = count;
+
+    while (remaining > 0) {
+	if ((bf->current_block) == (bf->idx->n)) { //
+	    fprintf(stderr, "ERROR: reached end of bgzip index with more data to write\n");
+	    return -1;
+	}
+	ublock_size = bf->idx->u_off[bf->current_block+1]
+	            - bf->idx->u_off[bf->current_block];
+
+	int copy_length = ublock_size - bf->bgbuf_sz;
+	if (copy_length > remaining)
+	    copy_length = remaining;
+
+	memcpy(bf->bgbuf_p + bf->bgbuf_sz, input, copy_length);
+	input += copy_length;
+	bf->bgbuf_sz += copy_length;
+	remaining -= copy_length;
+
+	if (bf->bgbuf_sz == ublock_size) {
+	    BGZF_WRITE(bf, level, bf->bgbuf_p, ublock_size);
+	    bf->bgbuf_sz=0;
+	    bf->current_block++;  // track the blocks
+	}
+    }
+
+    return 0;
+}
+
+
 static int bgzf_write(bam_file_t *bf, int level, const void *buf, size_t count) {
     unsigned char blk[Z_BUFF_SIZE+4];
     uint32_t len;
@@ -2925,14 +3047,6 @@ static int bgzf_flush(bam_file_t *bf) {
  * Returns 0 on success
  *        -1 on failure
  */
-#if defined(__GNUC__) && !(defined(__clang__) || defined(__ICC))
-// On later gcc releases the ALLOW_UAC code causes the vectorizor to
-// use SIMD instructions on unaligned memory access.  This is due to
-// our own abuse of char to int aliasing, but doing things the legal
-// way is still slower overall (5-14% depending on system and
-// compiler).
-__attribute__((optimize("no-tree-vectorize")))
-#endif
 int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
     char *auxh, aux_key[3], type;
     bam_aux_t val;
@@ -3126,7 +3240,7 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 		int n = b->len & ~1;
 		for (i = 0; i < n; i+=2) {
 #ifdef ALLOW_UAC
-		    *(int16_t *)cp = le_int2(code2base[*dat++]);
+		    *(int16_u *)cp = le_int2(code2base[*dat++]);
 		    cp += 2;
 #else
 		    cp[0] = "=ACMGRSVTWYHKDBN"[*dat >> 4];
@@ -3187,7 +3301,7 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 		    int n = b->len & ~3;
 		    for (; i < n; i+=4) {
 			//*cp++ = *dat++ + '!';
-			*(uint32_t *)cp = *(uint32_t *)dat + 0x21212121;
+			*(uint32_u *)cp = *(uint32_u *)dat + 0x21212121;
 			cp  += 4;
 			dat += 4;
 		    }
@@ -3389,12 +3503,12 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 	int i, n = bam_cigar_len(b);
 #endif
 
-#define CF_FLUSH()					\
-	do {						\
-	    if (BGZF_WRITE(fp, fp->level, fp->uncomp,	\
-			   fp->uncomp_p - fp->uncomp))	\
-		return -1;				\
-	    fp->uncomp_p=fp->uncomp;			\
+#define CF_FLUSH()						\
+	do {							\
+	    if (bgzf_block_write(fp, fp->level, fp->uncomp,	\
+				 fp->uncomp_p - fp->uncomp))	\
+		return -1;					\
+	    fp->uncomp_p=fp->uncomp;				\
 	} while(0)
 
 	/* If big endian, byte swap inline, write it out, and byte swap back */
@@ -3544,7 +3658,7 @@ int bam_write_header(bam_file_t *out) {
 
 	while (len) {
 	    int sz = BGZF_BUFF_SIZE < len ? BGZF_BUFF_SIZE : len;
-	    if (BGZF_WRITE(out, out->level, cp, sz))
+	    if (bgzf_block_write(out, out->level, cp, sz))
 		return -1;
 	    cp  += sz;
 	    len -= sz;
@@ -3596,6 +3710,16 @@ int bam_set_voption(bam_file_t *fd, enum bam_option opt, va_list args) {
     case BAM_OPT_BINNING:
 	fd->binning = va_arg(args, int);
 	break;
+
+    case BAM_OPT_IGNORE_CHKSUM:
+	fd->ignore_chksum = va_arg(args, int);
+	break;
+    case BAM_OPT_WITH_BGZIP_IDX:
+        fd->idx =  va_arg(args, gzi *);
+	break;
+    case BAM_OPT_OUTPUT_BGZIP_IDX:
+        fd->idx_fn =  va_arg(args, char *);
+	break;
     }
 
     return 0;
