@@ -118,6 +118,9 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
     {
 	HashData hd;
 
+	if (h->preservation_map)
+	    HashTableDestroy(h->preservation_map, 0);
+	    
 	if (!(h->preservation_map = HashTableCreate(4, HASH_NONVOLATILE_KEYS)))
 	    return NULL;
 
@@ -760,8 +763,10 @@ static int cram_compress_slice(cram_fd *fd, cram_container *c, cram_slice *s) {
     if (fd->use_bz2)
 	method |= 1<<BZIP2;
 
-    if (fd->use_rans)
-	method |= (1<<RANS0) | (1<<RANS1);
+    if (fd->use_rans) {
+	method  |= (1<<RANS0) | (1<<RANS1);
+	methodF |= (1<<RANS0) | (1<<RANS1);
+    }
 
     if (fd->use_lzma)
 	method |= (1<<LZMA);
@@ -1059,6 +1064,12 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 	    return -1;
     }
 
+    if (fd->unsorted == 2) {
+	if (fd->ref_lock) pthread_mutex_lock(fd->ref_lock);
+	fd->unsorted = 1;
+	if (fd->ref_lock) pthread_mutex_unlock(fd->ref_lock);
+    }
+
     return r ? -1 : 0;
 }
 
@@ -1347,7 +1358,7 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 	} else {
 	    s->hdr->ref_seq_id    = c->ref_id;
 	    s->hdr->ref_seq_start = first_base;
-	    s->hdr->ref_seq_span  = last_base - first_base + 1;
+	    s->hdr->ref_seq_span  = MAX(0, last_base - first_base + 1);
 	}
 	s->hdr->num_records = r2;
 
@@ -2323,7 +2334,7 @@ void cram_update_curr_slice(cram_container *c) {
     } else {
 	s->hdr->ref_seq_id    = c->curr_ref;
 	s->hdr->ref_seq_start = c->first_base;
-	s->hdr->ref_seq_span  = c->last_base - c->first_base + 1;
+	s->hdr->ref_seq_span  = MAX(0, c->last_base - c->first_base + 1);
     }
     s->hdr->num_records   = c->curr_rec;
 
@@ -3059,6 +3070,11 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	fd->last_slice = curr_rec - slice_rec;
 	c->slice_rec = c->curr_rec;
 
+	if (c->refs_used && bam_ref(b) >= 0 && bam_ref(b) >= fd->refs->nref) {
+	    fprintf(stderr, "Reference absent in header. Failing\n");
+	    return -1;
+	}
+
 	// Have we seen this reference before?
 	if (bam_ref(b) >= 0 && curr_ref >= 0 && bam_ref(b) != curr_ref && !fd->embed_ref &&
 	    !fd->unsorted && multi_seq) {
@@ -3072,7 +3088,7 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	    } else if (c->refs_used && c->refs_used[bam_ref(b)]) {
 		fprintf(stderr, "Unsorted mode enabled\n");
 		if (fd->ref_lock) pthread_mutex_lock(fd->ref_lock);
-		fd->unsorted = 1;
+		fd->unsorted = 2; // 2 is marker to reset block metrics stats
 		if (fd->ref_lock) pthread_mutex_unlock(fd->ref_lock);
 		fd->multi_seq = 1;
 	    }
