@@ -58,6 +58,7 @@ extern "C" {
 #include "io_lib/hash_table.h"       // From io_lib aka staden-read
 #include "io_lib/thread_pool.h"
 #include "io_lib/mFILE.h"
+#include "io_lib/bgzip.h"
 
 #ifdef SAMTOOLS
 // From within samtools/HTSlib
@@ -259,6 +260,9 @@ typedef struct {
 
     // To aid compression
     cram_metrics *m; // used to track aux block compression only
+
+    int crc32_checked;
+    uint32_t crc_part;
 } cram_block;
 
 struct cram_codec; /* defined in cram_codecs.h */
@@ -586,6 +590,9 @@ typedef struct cram_slice {
     // For going from BAM to CRAM; an array of auxiliary blocks per type
     int naux_block;
     cram_block **aux_block;
+
+    // Cache of converted BAM structs
+    bam_seq_t **bl;
 } cram_slice;
 
 /*-----------------------------------------------------------------------------
@@ -613,7 +620,7 @@ typedef struct {
     int nref;              // number of ref_entry
 
     char *fn;              // current file opened
-    FILE *fp;              // and the FILE* to go with it.
+    bzi_FILE *fp;          // and the bzi_FILE* to go with it.
 
     int count;             // how many cram_fd sharing this refs struct
 
@@ -943,7 +950,9 @@ enum cram_option {
     CRAM_OPT_BASES_PER_SLICE,
     CRAM_OPT_LOSSY_READ_NAMES,
     CRAM_OPT_PRESERVE_AUX_ORDER,
-    CRAM_OPT_PRESERVE_AUX_SIZE
+    CRAM_OPT_PRESERVE_AUX_SIZE,
+    CRAM_OPT_WITH_BGZIP_INDEX,
+    CRAM_OPT_OUTPUT_BGZIP_IDX
 };
 
 /* BF bitfields */
