@@ -27,11 +27,17 @@ extern "C" {
  * processor type too.
  */
 
-/* Set by autoconf */
+/* Set by autoconf.
+ *
+ * If you are attempting to compile without autoconf, copy this file to
+ * os.h and remove the @...@ lines below.
+ */
+#ifndef NO_AUTOCONF
 #define SP_LITTLE_ENDIAN
 
 #define HAVE_STDIO_EXT_H
 #define CRAM_IO_CUSTOM_BUFFERING
+#endif
 
 /* Mac FAT binaries or unknown. Auto detect based on CPU type */
 #if !defined(SP_BIG_ENDIAN) && !defined(SP_LITTLE_ENDIAN)
@@ -262,6 +268,9 @@ extern "C" {
 #define popen _popen
 #define pclose _pclose
 typedef int mode_t;
+#ifdef ftruncate
+#  undef ftruncate
+#endif
 #define ftruncate(fd,len) _chsize(fd,len)
 #define sysconf(x) 512
 #define NOPIPE
@@ -269,6 +278,13 @@ typedef int mode_t;
 #define NOSTRCASECMP
 #define NO_STRPTIME
 #undef HAVE_SYS_WAIT_H
+
+//Visual Studio on Windows does not like "static inline"
+//http://lists.cairographics.org/archives/cairo/2008-January/012726.html
+#define inline __inline
+
+//Visual Studio does not define ssize_t
+#define ssize_t long
 #endif
 
 
@@ -386,7 +402,9 @@ typedef unsigned int	uint_4;
  * Some handy definitions.
  */
 
+#ifndef _MSC_VER
 #define MAXINT8 (9223372036854775807LL)
+#endif
 #define MAXINT4 (INT_MAX)
 #define MAXINT2 (SHRT_MAX)
 
