@@ -215,7 +215,7 @@ int main(int argc, char **argv) {
 	printf("\nContainer pos %"PRId64" size %d\n", (int64_t)pos, c->length);
 	printf("    Ref id:            %d\n", c->ref_seq_id);
 	printf("    Ref pos:           %d + %d\n", c->ref_seq_start, c->ref_seq_span);
-	printf("    Rec counter:       %d\n", c->record_counter);
+	printf("    Rec counter:       %"PRId64"\n", c->record_counter);
        	printf("    No. recs:          %d\n", c->num_records);
 	printf("    No. bases          %"PRId64"\n", c->num_bases);
 	printf("    No. blocks:        %d\n", c->num_blocks);
@@ -252,6 +252,10 @@ int main(int argc, char **argv) {
 	printf("        T: %.4s\n", c->comp_hdr->substitution_matrix[3]);
 	printf("        N: %.4s\n", c->comp_hdr->substitution_matrix[4]);
 
+	printf("      TD map:\n");
+	for (i = 0; i < c->comp_hdr->nTL; i++)
+	    printf("        %3d: %s\n", i, c->comp_hdr->TL[i]);
+
 	printf("\n      Record encoding map:\n");
 	DumpMap2(c->comp_hdr->rec_encoding_map, stdout, "\t", 
 		 (char *)c->comp_hdr_block->data, ds_h);
@@ -282,7 +286,7 @@ int main(int argc, char **argv) {
 		    printf("%02x", s->hdr->md5[i]);
 		putchar('\n');
 	    }
-	    printf("\tRec counter      %d\n", s->hdr->record_counter);
+	    printf("\tRec counter      %"PRId64"\n", s->hdr->record_counter);
 	    printf("\tNo. records      %d\n", s->hdr->num_records);
 	    printf("\tNo. blocks       %d\n", s->hdr->num_blocks);
 	    printf("\tBlk IDS:         {");
@@ -715,7 +719,7 @@ int main(int argc, char **argv) {
 	    }
 
 	    for (id = 0; id < s->hdr->num_blocks; id++) {
-		HashData hd;
+		HashData hd = {0};
 		cram_block *b = s->block[id];
 		printf("\n\tBlock %d/%d\n", id+1, s->hdr->num_blocks);
 		printf("\t    Size:         %d comp / %d uncomp\n",
@@ -811,11 +815,11 @@ int main(int argc, char **argv) {
 
 	    k = (intptr_t) hi->key;
 	    if (k == -1) {
-		printf("Block CORE          , total size %10ld\n", hi->data.i);
+		printf("Block CORE              , total size %11"PRId64"\n", hi->data.i);
 		continue;
 	    }
 
-	    printf("Block content_id %3d, total size %10ld ",
+	    printf("Block content_id %7d, total size %11"PRId64" ",
 		   (int) k, hi->data.i);
 
 	    struct {
