@@ -119,18 +119,21 @@ static void usage(FILE *fp) {
     fprintf(fp, "    -Z             [Cram] Also compress using lzma.\n");
 #endif
     fprintf(fp, "    -n             [Cram] Discard read names where possible.\n");
-    fprintf(fp, "    -P             [Cram EXPERIMENTAL] Preserve all aux tags (incl RG,NM,MD)\n");
-    fprintf(fp, "    -p             [Cram EXPERIMENTAL] Preserve aux tag sizes ('i', 's', 'c')\n");
+    fprintf(fp, "    -P             Preserve all aux tags (incl RG,NM,MD)\n");
+    fprintf(fp, "    -p             Preserve aux tag sizes ('i', 's', 'c')\n");
+    fprintf(fp, "    -q             Don't add scramble @PG header line\n");
     fprintf(fp, "    -N integer     Stop decoding after 'integer' sequences\n");
     fprintf(fp, "    -t N           Use N threads (availability varies by format)\n");
     fprintf(fp, "    -B             Enable Illumina 8 quality-binning system (lossy)\n");
     fprintf(fp, "    -!             Disable all checking of checksums\n");
+    fprintf(fp, "    -g FILE        Convert to Bam using index (file.gzi)\n");
+    fprintf(fp, "    -G FILE        Output Bam index when bam input(file.gzi)\n");
 }
 
 int main(int argc, char **argv) {
     scram_fd *in, *out;
     bam_seq_t *s;
-    char imode[10], *in_f = "", omode[10], *out_f = "";
+    char imode[10], *in_f = "", omode[10], *out_f = "", *index_fn = NULL, *index_out_fn = NULL;
     int level = '\0'; // nul terminate string => auto level
     int c, verbose = 0;
     int s_opt = 0, S_opt = 0, embed_ref = 0, ignore_md5 = 0, decode_md = 0;
@@ -141,6 +144,7 @@ int main(int argc, char **argv) {
     refs_t *refs;
     int nthreads = 1;
     t_pool *p = NULL;
+    gzi *idx =NULL;
     int max_reads = -1;
     enum quality_binning binning = BINNING_NONE;
     int sam_fields = 0; // all
@@ -148,10 +152,13 @@ int main(int argc, char **argv) {
     int bases_per_slice = 0;
     int lossy_read_names = 0;
     int preserve_aux_order = 0;
-    int preserve_aux_size = 0;
+    int preserve_aux_size = 0; 
+    int add_pg = 1;   
+
+    scram_init();
 
     /* Parse command line arguments */
-    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xXeI:O:R:!MmjJZt:BN:F:Hb:nPp")) != -1) {
+    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xXeI:O:R:!MmjJZt:BN:F:Hb:nPpqg:G:")) != -1) {
 	switch (c) {
 	case 'F':
 	    sam_fields = strtol(optarg, NULL, 0); // undocumented for testing
@@ -298,10 +305,22 @@ int main(int argc, char **argv) {
 	    preserve_aux_size = 1;
 	    break;
 
+	case 'q':
+	    add_pg = 0;
+	    break;
+
 	case 'N':
 	    max_reads = atoi(optarg);
 	    break;
 
+	case 'g':
+	    index_fn = optarg;
+	    break;
+
+	case 'G':
+	    index_out_fn = optarg;
+	    break;
+
 	case '?':
 	    fprintf(stderr, "Unrecognised option: -%c\n", optopt);
 	    usage(stderr);
@@ -409,6 +428,20 @@ int main(int argc, char **argv) {
 	    return 1;
     }
 
+    if (index_fn) {
+	if (NULL == (idx = gzi_index_load(index_fn))) {
+	    fprintf(stderr, "Cannot open index file.\n");
+	    return 1;
+	}
+	if (scram_set_option(out, CRAM_OPT_WITH_BGZIP_INDEX, idx))
+	    return 1;
+    }
+
+    if (index_out_fn) {
+	if (scram_set_option(in, CRAM_OPT_OUTPUT_BGZIP_IDX, index_out_fn))
+	    return 1;
+    }
+
     if (nthreads > 1) {
 	if (NULL == (p = t_pool_init(nthreads*2, nthreads)))
 	    return 1;
@@ -457,21 +490,23 @@ int main(int argc, char **argv) {
     }
 
     if (scram_get_header(out)) {
-	char *arg_list = stringify_argv(argc, argv);
+        if (add_pg) {
+	    char *arg_list = stringify_argv(argc, argv);
 
-	if (!arg_list)
-	    return 1;
+	    if (!arg_list)
+		return 1;
 
-	if (sam_hdr_add_PG(scram_get_header(out), "scramble",
-			   "VN", PACKAGE_VERSION,
-			   "CL", arg_list, NULL))
-	    return 1;
+	
+	    if (sam_hdr_add_PG(scram_get_header(out), "scramble",
+			       "VN", PACKAGE_VERSION,
+			       "CL", arg_list, NULL))
+	        return 1;
+
+	    free(arg_list);
+	}
 
 	if ((header || omode[1] != 's') && scram_write_header(out) != 0)
 	    return 1;
-
-	free(arg_list);
-
     }
 
 
